#include <stdio.h>
#include <stdlib.h>
#include <string.h>

unsigned char *buf;
int buflen, tunen;

void savetune( unsigned char *buf, char *name, char *ext, int tunelen, int zeropad )
{
  char wname[256];
  FILE *f;
  int j;

  tunen++;
  if( name )
  {
    strncpy( wname, name, 120 );
    if( ( strlen( name ) < 4 ) ||
        ( stricmp( &name[strlen(name)-4], ext ) != 0 ) )
      strcat( wname, ext );
  } else {
    sprintf( wname, "%05d%s", tunen, ext );
  }
  f = fopen( wname, "wb" );
  if( f )
  {
    fwrite( buf, tunelen, 1, f );
    if( zeropad > 0 )
    {
      for( j=0; j<256; j++ ) wname[j] = 0;
      while( zeropad > 0 )
      {
        j = zeropad<256 ? zeropad : 256;
        fwrite( wname, j, 1, f );
        zeropad -= j;
      }
    }
    fclose( f );
  }
}

int find_ahx( int i, int *zp, char **name )
{
  unsigned char *b, *bptr;
  int j, k, l, posn, insn, ssn, trkn, trkl;

  *name = NULL;

  // Enough space left for an AHX tune?
  if( ( buflen-i ) < 14 ) return 0;

  // Header
  if( strncmp( (char *)&buf[i], "THX", 3 ) != 0 )
    return 0;
  if( buf[i+3] >= 3 )
    return 0;

  b = &buf[i];

  posn = ((b[6]&0x0f)<<8)|b[7];
  insn = b[12];
  ssn  = b[13];
  trkl = b[10];
  trkn = b[11];

  if( ( posn > 1000 ) ||
      ( trkl > 64 ) ||
      ( insn > 64 ) )
  {
    printf( "Possible corrupt tune at %d (PosN=%d,TrackLen=%d,Instruments=%d)?\n",
      i, posn, trkl, insn );
    return 0;
  }

  /* Verify that the song data seems valid */
  j = 14;
  j += ssn*2;    // Skip past the subsong list
  j += posn*4*2; // Skip past the positions
  j += trkn*trkl*3;
  if((b[6]&0x80)==0) j += trkl*3;

  for( k=1; k<=insn; k++ )
  {
    if( (j+i+22) >= buflen )
    {
      printf( "Possible truncated AHX tune at end of file?\n" );
      return 0;
    }

    j += 22 + b[j+21]*4;
  }

  if( (j+i) >= buflen )
  {
    printf( "Possible truncated AHX tune at end of file?\n" );
    return 0;
  }

  /* Check the text stuff (sometimes this gets chopped off but its not a problem */
  k = (b[4]<<8)|b[5];

  if( k < j )
  {
    printf( "Potential AHX tune detected with invalid strings pointer.\n" );
    printf( "Saving with empty and song titles; the file might be corrupt.\n" );
    printf( "(tune %d)\n", tunen+1 );
    buf[4] = (j>>8)&0xff;
    buf[5] = j&0xff;
    savetune( b, NULL, ".ahx", j, insn+1 );
    buf[4] = (k>>8)&0xff;
    buf[5] = k&0xff;
    return 0;
  }

  // Check for correct number of null-terminated strings
  l=0;
  while( l < (insn+1) )
  {
    j=0;
    while( (j+k+i) < buflen )
    {
      if( b[j+k] == 0 ) break;
      if( ( b[j+k] < 32 ) || ( b[j+k] > 127 ) )  // Try and get only ASCII chars in strings
        break;
      j++;
    }

    if( (j+k+i) >= buflen ) break;
    if( b[j+k] != 0 ) break;

    // Found a NULL terminated string!
    if( ( l == 0 ) && ( j > 0 ) )
      *name = &b[k];

    l++;
    k+=j+1;
  }

  *zp = (insn+1)-l;
  return k;
}

int find_hvl( int i, int *zp, char **name )
{
  unsigned char *b, *bptr;
  int j, k, l, chnn, posn, insn, ssn, trkn, trkl;

  *name = NULL;

  // Enough space left for an HVL tune?
  if( ( buflen-i ) < 16 ) return 0;

  // Header
  if( strncmp( (char *)&buf[i], "HVL", 3 ) != 0 )
    return 0;
  if( buf[i+3] > 0 )
    return 0;

  b = &buf[i];

  posn = ((b[6]&0x0f)<<8)|b[7];
  insn = b[12];
  ssn  = b[13];
  chnn = (b[8]>>2)+4;
  trkl = b[10];
  trkn = b[11];

  if( ( posn > 1000 ) ||
      ( trkl > 64 ) ||
      ( insn > 64 ) ||
      ( chnn > 16 ) )
  {
    printf( "Possible corrupt tune at %d (PosN=%d,TrackLen=%d,Instruments=%d,Channels=%d)?\n",
      i, posn, trkl, insn, chnn );
    return 0;
  }

  j = 16;
  j += ssn*2;    // Skip past the subsong list
  j += posn*chnn*2; // Skip past the positions

  // Skip past the tracks
  for( k=((b[6]&0x80)==0x80)?1:0; k<=trkn; k++ )
    for( l=0; l<trkl; l++ )
    {
      if( (j+i) >= buflen )
      {
        printf( "Possible truncated HVL tune at end of file?\n" );
        return 0;
      }

      if( b[j] == 0x3f )
      {
        j++;
        continue;
      }

      j += 5;
    }

  for( k=1; k<=insn; k++ )
  {
    if( (j+i+22) >= buflen )
    {
      printf( "Possible truncated HVL tune at end of file?\n" );
      return 0;
    }

    j += 22 + b[j+21]*5;
  }

  if( (j+i) >= buflen )
  {
    printf( "Possible truncated HVL tune at end of file?\n" );
    return 0;
  }

  /* Check the text stuff (sometimes this gets chopped off but its not a problem */
  k = (b[4]<<8)|b[5];

  if( k < j )
  {
    printf( "Potential HVL tune detected with invalid strings pointer.\n" );
    printf( "Saving with empty and song titles; the file might be corrupt.\n" );
    printf( "(tune %d)\n", tunen+1 );
    buf[4] = (j>>8)&0xff;
    buf[5] = j&0xff;
    savetune( b, NULL, ".hvl", j, insn+1 );
    buf[4] = (k>>8)&0xff;
    buf[5] = k&0xff;
    return 0;
  }

  // Check for correct number of null-terminated strings
  l=0;
  while( l < (insn+1) )
  {
    j=0;
    while( (j+k+i) < buflen )
    {
      if( b[j+k] == 0 ) break;
      if( ( b[j+k] < 32 ) || ( b[j+k] > 127 ) )  // Try and get only ASCII chars in strings
        break;
      j++;
    }

    if( (j+k+i) >= buflen ) break;
    if( b[j+k] != 0 ) break;

    // Found a NULL terminated string!
    if( ( l == 0 ) && ( j > 0 ) )
      *name = &b[k];

    l++;
    k+=j+1;
  }

  *zp = (insn+1)-l;
  return k;
}

int main(int argc, char *argv[])
{
  FILE *f;
  char *name;
  int i, tunelen, zeropad;

  printf( "AHX and HVL Ripper v1.0\nBy Xeron/IRIS\n" );

  if( argc < 2 )
  {
    printf( "I need a file to rip stuff out of!!!1\n" );
    return 0;
  }

  f = fopen( argv[1], "rb" );
  if( !f )
  {
    printf( "Can't open '%s'\n", argv[1] );
    return 0;
  }

  fseek( f, 0, SEEK_END );
  buflen = ftell( f );
  fseek( f, 0, SEEK_SET );

  buf = malloc( buflen );
  if( !buf )
  {
    printf( "Oh noes! I can't allocate %d bytes!\n", buflen );
    fclose( f );
    return 0;
  }

  fread( buf, buflen, 1, f );
  fclose( f );

  // Lets do it!
  i=0;
  tunen=0;
  while( i < buflen )
  {
    tunelen = find_ahx( i, &zeropad, &name );
    if( tunelen > 0 )
    {
      printf( "Found AHX tune at %d (size %d) [%s]\n", i, tunelen+zeropad, name ? name : "untitled" );
      savetune( &buf[i], name, ".ahx", tunelen, zeropad );
      i += 4;      
      continue;
    }

    tunelen = find_hvl( i, &zeropad, &name );
    if( tunelen > 0 )
    {
      printf( "Found HVL tune at %d (size %d) [%s]\n", i, tunelen+zeropad, name ? name : "untitled" );
      savetune( &buf[i], name, ".hvl", tunelen, zeropad );
      i += 4;
      continue;
    }

    i++;
  }

  if( tunen == 0 )
  {
    printf( "No AHX or HVL tunes found. Maybe the file is packed?\n" );
    return 0;
  }

  return 0;
}
